require "debugprint" local Widget = require "widgets/widget"
local Text = require "widgets/text"
local ImageButton = require "widgets/imagebutton"
local UIAnim = require "widgets/uianim"


local FastForward = Class(Widget, function(self, owner)
  Widget._ctor(self, "FastForward", owner)

  self.max_scale = 4

  self.target_scale = 1
  self.stop_tick = 0
  self.force_text_mark = 0

  --subwidgets
  self.anim = self:AddChild(UIAnim())
  local anim = self.anim:GetAnimState()
  anim:SetBuild("rotating")
  anim:SetBank("rotating_light")
  anim:SetMultColour(0,1,0,1)
  anim:SetScale(.6,.6,.6)
  anim:SetDeltaTimeMultiplier(.4)
  anim:PlayAnimation("rotating_on", true)

  self.gauge = self:AddChild(Image("images/hud.xml", "clock_rim.tex"))
  self.gauge.owner = self
  self.gauge:SetScale(.6,.6,.6)
  self.gauge:SetTooltip(STRINGS.UI.HUD.SPEEDWHEEL)


  self.text = self:AddChild(Text(BODYTEXTFONT, 33))
  self.text.owner = self
  self.text:SetHAlign(ANCHOR_MIDDLE)
  self.text:SetPosition(5,0,0)
  self.text:Show()

  self.minus_button = self:AddChild(ImageButton("images/button_minus.xml", "button_minus.tex"))
  self.minus_button:SetPosition(-45,0,0)
  self.minus_button:SetScale(.4,.4,.4)
  self.minus_button:SetOnClick(function() self:ChangeSpeed(-.5) end)
  self.minus_button:SetTooltip(STRINGS.UI.HUD.SPEEDDOWN)	

  self.plus_button = self:AddChild(ImageButton("images/button_plus.xml", "button_plus.tex"))
  self.plus_button:SetPosition(45,0,0)
  self.plus_button:SetScale(.4,.4,.4)
  self.plus_button:SetOnClick(function() self:ChangeSpeed(.5) end)
  self.plus_button:SetTooltip(STRINGS.UI.HUD.SPEEDUP)	

  self:UpdateText()
  self:StartUpdating()

  --handle wheel
  local gauge_oncontrol = self.gauge.OnControl
  self.gauge.OnControl = function(self, control)

    if self.focus or TheInput:IsKeyDown(self.parent.togglekey) then
      if control == CONTROL_MAP_ZOOM_IN or control == CONTROL_ZOOM_IN then
	self.owner.force_text_mark = GetTime()
	self.owner:ChangeSpeed(.1) 
      elseif control == CONTROL_MAP_ZOOM_OUT or control == CONTROL_ZOOM_OUT then
	self.owner.force_text_mark = GetTime()
	self.owner:ChangeSpeed(-.1) 
      else
	return gauge_oncontrol(self, control)
      end
      return true
    end
  end

  --intercept camera events, maintain its behaviour
  local controller = GetPlayer().components.playercontroller
  controller.DoCameraControl_original = controller.DoCameraControl
  controller.fast_forward = self

  controller.DoCameraControl = function(self)
    if (TheInput:IsControlPressed(CONTROL_ZOOM_IN) or 
      TheInput:IsControlPressed(CONTROL_ZOOM_OUT)) then
      if self.fast_forward.focus then
	return 
      elseif self.lastzoomtime then
	local elapsed = GetTime() - self.lastzoomtime
	if elapsed / TheSim:GetTimeScale() < .1 then return end
      end
    end
    if TheInput:IsControlPressed(CONTROL_ROTATE_LEFT) or
      TheInput:IsControlPressed(CONTROL_ROTATE_RIGHT) then
      if TheInput:IsKeyDown(self.fast_forward.togglekey) then
	self.fast_forward.force_text_mark = GetTime()
	if TheInput:IsControlPressed(CONTROL_ROTATE_LEFT) then
	  self.fast_forward:ChangeSpeed(-.5)
	else
	  self.fast_forward:ChangeSpeed(.5)
	end
	return
      elseif self.lastrottime then
	local elapsed = GetTime() - self.lastrottime
	if elapsed / TheSim:GetTimeScale() < .25 then return end
      end
    end
    self:DoCameraControl_original(self)
  end
end)

function FastForward:OnUpdate(dt)

  if IsPaused() then return end

  local in_effect_scale = TheSim:GetTimeScale()
  local candidate_scale = in_effect_scale

  local controller = GetPlayer().components.playercontroller
  local slow_down = controller.inst.sg:HasStateTag("idle")

  if not slow_down then
    local in_danger = METRICS_ENABLED and GetPlayer().components.overseer.engaged
    local in_darkness = GetPlayer().LightWatcher:GetLightValue() <= 
      TUNING.SANITY_LOW_LIGHT
    local is_weak = GetPlayer().components.hunger.current < 20 or
      GetPlayer().components.health.currenthealth < 30
    local is_crazy = GetPlayer().components.sanity.current < 30

    if (self.mode < 3) then
      slow_down = in_danger or in_darkness or is_weak
      if (self.mode == 1) then slow_down = slow_down or is_crazy; end
    else
      slow_down = not is_crazy and (in_danger or in_darkness or is_weak)
    end
  end

  if not slow_down and controller.inst.bufferedaction and
    controller.inst.bufferedaction.action == ACTIONS.CHOP then
    local x, y, z = GetPlayer().Transform:GetWorldPosition();
    local ents = 
      TheSim:FindEntities(x, y, z, TUNING.LEIF_MAXSPAWNDIST+1, {"leif"})
    if #ents > 0 then
      slow_down = true
    end
  end

  if slow_down then
    candidate_scale = math.max(1, in_effect_scale - 8*dt)
  else
    candidate_scale = math.min(self.target_scale, in_effect_scale + dt)
  end

  if candidate_scale ~= in_effect_scale then
    TheSim:SetTimeScale(candidate_scale)

    local lin = (candidate_scale - 1) / (self.max_scale - 1)
    local r = math.min(1, 2 * lin)
    local g = math.min(1, 2 * (1-lin))
    local anim = self.anim:GetAnimState()
    anim:SetMultColour(r,g,0,1)
    anim:SetDeltaTimeMultiplier(.4 + lin * 1.2)

    self:UpdateText()
  elseif self.force_text_mark > 0 then
    self:UpdateText()
  end
end


function FastForward:OnGainFocus()
  self:UpdateText()
end

function FastForward:OnLoseFocus()
  self.force_text_mark = 0
  self:UpdateText()
end

function FastForward:ChangeSpeed(inc)
  self.target_scale = math.max(1,math.min(self.max_scale, 
    self.target_scale + inc))
  self:UpdateText()
end

function FastForward:UpdateText(force)
  local force = false
  if self.force_text_mark > 0 and
    (GetTime() - self.force_text_mark) / TheSim:GetTimeScale() < .5 then
    force = true
  else
    self.force_text_mark = 0
  end
  if self.focus or force then
    self.text:SetSize(22)
    self.text:SetString(string.format("max\nx%.1f", self.target_scale))
  else
    self.text:SetSize(33)
    self.text:SetString(string.format("x%.1f", TheSim:GetTimeScale()))
  end
end

return FastForward
